package ga.core;

import ga.classification.Classifier;

import java.io.Serializable;

/**
 *
 * An individual has a set of parameters which are to be tuned.
 *
 * @author Olly Oechsle, University of Essex, Date: 18-May-2007
 * @version 1.0
 */
public abstract class GAIndividual implements Comparable, Serializable {

    public GAGene[] genes;
    protected double kozaFitness;
    public double alternativeFitness;
    protected int hits;
    protected Classifier classifier;
    public boolean evaluated = false;

    public GAIndividual(int size) {
        genes = new GAGene[size];
    }

    public Classifier getClassifier() {
        return classifier;
    }

    public void setClassifier(Classifier classifier) {
        this.classifier = classifier;
    }

    public double getKozaFitness() {
        return kozaFitness;
    }

    public void setKozaFitness(double kozafitness) {
        this.kozaFitness = kozafitness;
    }

    public void setHits(int hits) {
        this.hits = hits;
    }

    public int getHits() {
        return hits;
    }

    public void setWorstFitness() {
        this.kozaFitness = Double.MAX_VALUE;
    }

    public int compareTo(Object o) {
        GAIndividual other = (GAIndividual) o;
        if (other.kozaFitness == this.kozaFitness) return 0;
        if (other.kozaFitness < this.getKozaFitness()) return +1;
        return -1;
    }

    /**
     * Returns the raw value after executing the associated function whose parameters
     * are defined by this individual.
     */
    public abstract double execute(Object o);

    /**
     * Prints out the individual as a line of java code.
     * @return
     */
    public String toJava() {
        StringBuffer buffer = new StringBuffer("new double[] {");
        for (int i = 0; i < genes.length; i++) {
            GAGene gene = genes[i];
            buffer.append(gene);
            if (i < genes.length - 1) buffer.append(", ");
        }
        buffer.append("];");
        return buffer.toString();
    }

    public int classify(Object o) {

        if (classifier != null) {
            return classifier.classify(execute(o));
        } else {
            throw new RuntimeException("No classifier assigned to individual.");
        }

    }

    public String toString() {
        StringBuffer buffer = new StringBuffer("");
        for (int i = 0; i < genes.length; i++) {
            GAGene gene = genes[i];
            buffer.append(gene);
            if (i < genes.length - 1) buffer.append(" | ");
        }
        return buffer.toString();
    }

}
