package problems.boosting;


import ga.core.GAProblemBoostable;
import ga.core.GAIndividual;
import ga.Evolve;

import ga.listeners.GAConsoleListener;
import problems.face.FaceDetectorProblem1;
import ac.essex.ooechs.adaboost.AdaBoostLearner;
import ac.essex.ooechs.adaboost.AdaBoostSample;
import ac.essex.ooechs.adaboost.AdaBoost;

import java.io.*;
import java.util.Vector;

/**
 * Wraps around another problem and applies AdaBoost.
 *
 * @author Olly Oechsle, University of Essex, Date: 21-May-2007
 * @version 1.0
 */
public class BoostingProblem extends AdaBoost {

    protected GAProblemBoostable problem;

    public void save(String className, String dir) throws Exception {
        File file = new File(new File(dir), className + ".java");
        FileWriter fileWriter = new FileWriter(file);
        BufferedWriter out = new BufferedWriter(fileWriter);
        String text = toJava(className);
        out.write(text);
        out.close();
    }

    public BoostingProblem(GAProblemBoostable p) {
        this.problem = p;
        individuals = new Vector<GAIndividual>(10);
    }

    public AdaBoostSample[] getSamples() {
        return problem.getSamples();
    }

    private Vector<GAIndividual> individuals;

    protected AdaBoostLearner weakLearn(AdaBoostSample[] samples, double[] weights) {

        problem.setWeights(weights);

        double lowestE = Double.MAX_VALUE;
        GAIndividual bestIndividual = null;

        for (int funcID = 1; funcID <= problem.countFunctions(); funcID++) {
            Evolve e = new Evolve(problem, new GAConsoleListener(true));
            problem.setFunctionID(funcID);
            e.run();
            GAIndividual ind = e.getBestIndividual();
            System.out.println("Individual TP: ( func" + funcID + " ) " + ind.getHits() + "/" + samples.length + ", e=" + ind.getKozaFitness());
            System.out.println(ind.toJava());
            if (ind.getKozaFitness() < lowestE) {
                lowestE = ind.getKozaFitness();
                bestIndividual = ind;
            }

        }

        System.out.println("Choosing: " + bestIndividual.getHits() + ", e=" + bestIndividual.getKozaFitness());

        individuals.add(bestIndividual);

        return new GAWeakLearner(bestIndividual);


    }

    public String toJava(String classname) {

        StringBuffer buffer = new StringBuffer();

        if (problem instanceof FaceDetectorProblem1) {
            buffer.append("package problems.face.solutions;\n\n");
        } else {
            buffer.append("package problems.jasmine.solutions.pasta;");
        }

        buffer.append("import ac.essex.ooechs.adaboost.AdaBoostSolution;\n" +
                "import ac.essex.ooechs.adaboost.AdaBoostLearner;\n" +
                "import ac.essex.ooechs.adaboost.AdaBoostSample;\n" +
                "import ac.essex.ooechs.ecj.haar.solutions.FaceDetector;\n" +
                "import ac.essex.ooechs.imaging.commons.HaarRegions;\n" +
                "import ac.essex.ooechs.ecj.commons.util.ObjectClass;\n" +
                "import ac.essex.ooechs.imaging.commons.apps.shapes.SegmentedShape;\n" +
                "import problems.jasmine.ShapeIndividual;" +
                "import ga.classification.BinaryClassifier;\n");

        buffer.append("public class ");
        buffer.append(classname);

        if (problem instanceof FaceDetectorProblem1) {
            buffer.append(" extends FaceDetector");
        }


        buffer.append("{\n\n");

        buffer.append("AdaBoostSolution tester;\n\n");

        // constructor
        buffer.append("public ");
        buffer.append(classname);
        buffer.append("() {\n");

        buffer.append("tester = new AdaBoostSolution(");
        buffer.append(individuals.size());
        buffer.append(");\n");

        for (int i = 0; i < individuals.size(); i++) {

            buffer.append("tester.addHypothesis(new WeakLearner");
            buffer.append(i);
            buffer.append("(), ");
            buffer.append(b[i]);
            buffer.append(");");

        }

        buffer.append("}\n\n");


        for (int i = 0; i < individuals.size(); i++) {
            GAIndividual gaIndividual = individuals.elementAt(i);

            buffer.append("class WeakLearner");
            buffer.append(i);
            buffer.append(" extends AdaBoostLearner {\n");

            buffer.append("public int classify(AdaBoostSample sample, Object data) {\n");

            if (problem instanceof FaceDetectorProblem1) {

                buffer.append("\tHaarRegions image = (HaarRegions) sample.getData();\n");
                buffer.append("\treturn ");
                buffer.append(gaIndividual.getClassifier().toJava(gaIndividual.toJava()));

            } else {

                buffer.append("\tSegmentedShape shape = (SegmentedShape) sample.getData();");
                buffer.append("\treturn ");
                buffer.append(gaIndividual.toJava());

            }

            buffer.append(";\n");

            buffer.append("}\n");

            buffer.append("}\n\n");

        }


        if (problem instanceof FaceDetectorProblem1) {

            buffer.append("public double calculate(HaarRegions haarRegions) {\n");

            buffer.append("return tester.classify(new AdaBoostSample(haarRegions)) == BinaryClassifier.TRUE? 1 : ObjectClass.NO_CLASS;\n");

            buffer.append("}\n");

        } else {

            buffer.append("public boolean classify(SegmentedShape shape) {\n");

            buffer.append("return tester.classify(new AdaBoostSample(shape)) == 1.0;\n");

            buffer.append("}\n");

        }


        buffer.append("}");

        return buffer.toString();


    }


}
