package problems.face;

import ac.essex.ooechs.imaging.commons.HaarRegions;
import ac.essex.ooechs.imaging.commons.PixelLoader;

import java.text.DecimalFormat;
import java.io.File;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.Vector;
import java.util.Collections;

import ga.Evolve;

/**
 * <p/>
 * Attempts to find the best detectors using brute force search.
 * </p>
 *
 * @author Olly Oechsle, University of Essex, Date: 31-Oct-2007
 * @version 1.0
 */
public class AutomaticHaarFeatureChooser {

    public static final int ONE_RECT = 1;
    public static final int TWO_RECT = 2;
    public static final int THREE_RECT = 3;
    public static final int FOUR_RECT = 4;

    public static int WIDTH = 19;
    public static int HEIGHT = 19;

    private static final int MAX_FILES = 500;

    Vector<HaarRegions> trueFiles, falseFiles;
    Vector<HaarFeature> features;

    int numberOfFeaturesRequired;

    public static void main(String[] args) throws Exception {

/*        File trainingTrue = new File("/home/ooechs/Data/faces/train/face");
        File trainingFalse = new File("/home/ooechs/Data/faces/train/non-face");

        File testingTrue = new File("/home/ooechs/Data/faces/test/face");
        File testingFalse = new File("/home/ooechs/Data/faces/test/non-face");

        AutomaticHaarFeatureChooser fp3 = new AutomaticHaarFeatureChooser();

        fp3.getClassifiers(10, trainingTrue, trainingFalse);

        fp3.toCSV("faces-auto.csv", trainingTrue, trainingFalse);
        fp3.toCSV("faces-auto-unseen.csv", testingTrue, testingFalse);*/

        WIDTH = 50;
        HEIGHT = 5;

        File trainingTrue = new File("/home/ooechs/Desktop/pipes/Foreground");
        File trainingFalse = new File("/home/ooechs/Desktop/pipes/Background");

        AutomaticHaarFeatureChooser fp3 = new AutomaticHaarFeatureChooser();

        fp3.getClassifiers(20, trainingTrue, trainingFalse);

        fp3.toCSV("pipe-features.csv", trainingTrue, trainingFalse);


    }

    public AutomaticHaarFeatureChooser() {
        features = new Vector<HaarFeature>(10);
    }

    public void toCSV(String filename, File trueDirectory, File falseDirectory) throws Exception {

        File f = new File("/home/ooechs/Desktop/" + filename);
        BufferedWriter out = new BufferedWriter(new FileWriter(f));

/*        features.add(new HaarFeature(TWO_RECT, 6, 5, 1, 8, 1, 1)); //59%
        features.add(new HaarFeature(TWO_RECT, 5, 0, 2, 4, 1, 0)); // 71-74%
        features.add(new HaarFeature(TWO_RECT, 0, 2, 16, 4, 2, 1));// 82-85%
        features.add(new HaarFeature(TWO_RECT, 4, 3, 1, 4, 2, 1)); // 83-85%
        features.add(new HaarFeature(TWO_RECT, 9, 0, 2, 6, 1, 1)); // 77-78%
        features.add(new HaarFeature(FOUR_RECT, 0, 0, 4, 2, 1, 0)); //77-80%
        features.add(new HaarFeature(TWO_RECT, 1, 3, 17, 3, 2, 1)); // 85-86%
        features.add(new HaarFeature(TWO_RECT, 14, 2, 3, 4, 2, 1)); // 82-84%
        features.add(new HaarFeature(FOUR_RECT, 7, 0, 6, 8, 1, 0)); // 76%
        features.add(new HaarFeature(FOUR_RECT, 8, 0, 1, 3, 1, 0)); // 78-79%*/

        toCSV(out, features, trueDirectory, 1);
        toCSV(out, features, falseDirectory, 0);

        out.close();

    }

    private void toCSV(BufferedWriter out, Vector<HaarFeature> features, File directory, int classID) throws Exception {

        File[] files = directory.listFiles();

        for (int i = 0; i < files.length; i++) {
            File f = files[i];
            if (f.isDirectory()) continue;
            HaarRegions image = new HaarRegions(new PixelLoader(f));
            image.makeWindowFillImage(WIDTH, HEIGHT);
            for (int j = 0; j < features.size(); j++) {
                HaarFeature feature = features.elementAt(j);
                out.write(String.valueOf(feature.execute(image)));
                out.write(",");
            }
            out.write(classID + "\n");
        }
    }

    public void getClassifiers(int numberOfFeaturesRequired, File trueDirectory, File falseDirectory) {
        this.numberOfFeaturesRequired = numberOfFeaturesRequired;
        for (int i = 0; i < 1; i++) {
            initVectors();
            load(trueFiles, trueDirectory);
            load(falseFiles, falseDirectory);
            findFeatures();
        }
        Collections.sort(features);
        System.out.println("---");
        for (int i = 0; i < features.size() && i < numberOfFeaturesRequired; i++) {
            HaarFeature haarFeature = features.elementAt(i);
            System.out.println(haarFeature.getHits());
            System.out.println(haarFeature.toJava());
        }

        System.out.println("----");

        for (int i = 0; i < featuresAtX.size(); i++) {
            HaarFeature feature = featuresAtX.elementAt(i);
            System.out.println(i + ", " + feature.getHits() + ", " + feature.toJava());
        }
        
    }

    private void initVectors() {
        trueFiles = new Vector<HaarRegions>();
        falseFiles = new Vector<HaarRegions>();
    }

    private void load(Vector<HaarRegions> images, File directory) {
        File[] files = directory.listFiles();
        while (images.size() < MAX_FILES) {
            File file = files[(int) (files.length * Evolve.getRandomNumber())];
            try {
                HaarRegions image = new HaarRegions(new PixelLoader(file));
                image.makeWindowFillImage(WIDTH, HEIGHT);
                images.add(image);
            } catch (Exception e) {
                System.err.println(e.getMessage());
                return;
            }
        }
        System.out.println("Loaded " + images.size() + " files.");
    }

    Vector<HaarFeature> featuresAtX;

    int MAX_WIDTH = 10;

    public void findFeatures() {

        featuresAtX = new Vector<HaarFeature>(10);


        DecimalFormat f = new DecimalFormat("0.00");

        double highest = 0;

        for (int x = 0; x < WIDTH; x++) {

            HaarFeature bestAtX = null;

            for (int y = 0; y < HEIGHT; y++) {

                for (int height = 1; height < (HEIGHT - y); height++) {
                    for (int width = 1; width < (WIDTH - x) && width < MAX_WIDTH; width++) {
                        if (width == 1 && height == 1) continue;
                        for (int mode = TWO_RECT; mode <= FOUR_RECT; mode++) {
                            for (int adjacency = HaarRegions.HORIZONTALLY_ADJACENT; adjacency <= HaarRegions.VERTICALLY_ADJACENT; adjacency++)
                            {
                                int shape = HaarRegions.FIRST_SHAPE;
                                try {

                                    HaarFeature feature = new HaarFeature(mode, x, y, width, height, adjacency, shape);

                                    double[] trueCache = feature.execute(trueFiles);
                                    double trueValue = feature.mean;
                                    double[] falseCache = feature.execute(falseFiles);
                                    double falseValue = feature.mean;

                                    // the classifier can't make any distinction
                                    if (trueValue == falseValue) continue;

                                    double threshold = (trueValue + falseValue) / 2;



                                    int hits = calculateHits(trueCache, falseCache, threshold);
                                    feature.setHits(hits);
                                    
 /*                                   //if (hits > highest) {

                                    //highest = hits;
                                    String code = feature.toJava();
                                    double percentage = hits / (double) (trueCache.length + falseCache.length);
                                    //System.out.println("hits: " + hits + "(" + f.format(percentage) + ")");
                                    //System.out.println(code);
                                    if (percentage > 0.65) {
                                        //  only add if above this goodness threshold
                                        //addFeature(feature);
                                    }*/

                                    if (bestAtX == null || bestAtX.getHits() < feature.getHits()) {
                                        bestAtX = feature;
                                    }
                                    //}

                                } catch (Exception e) {
                                    // sometimes with or height is too much - ignore
                                }
                            }
                        }
                    }
                }               
            }

            System.out.println("X: " + x);
            System.out.println(bestAtX);
            //featuresAtX.add(bestAtX);
            if (bestAtX != null)
            features.add(bestAtX);

        }
    }

    private void addFeature(HaarFeature feature) {
        if (features.size() < (numberOfFeaturesRequired * 2)) {
            features.add(feature);
            //System.out.println("Adding classifier: HITS:" + feature.getHits());
        } else {
            HaarFeature similar = null;
            double mostSimilar = Double.MAX_VALUE;
            // find which existing feature this feature is most similar to
            for (int i = 0; i < features.size(); i++) {
                HaarFeature other = features.elementAt(i);
                double similarity = feature.estimateSimilarity(other);
                if (similar == null || similarity < mostSimilar) {
                    mostSimilar = similarity;
                    similar = other;
                }
            }
            // if i am better than the other one, replace it with me
            if (feature.getHits() > similar.getHits()) {
                //System.out.println("Replacing classifier " + similar.getHits() + " with " + feature.getHits());
                features.remove(similar);
                features.add(feature);
            }

        }
    }

    public void getHits(int mode, int x, int y, int width, int height, int adjacency, int shape) {
        getHits(new HaarFeature(mode, x, y, width, height, adjacency, shape));
    }

    public void getHits(HaarFeature feature) {

        double[] trueCache = feature.execute(trueFiles);
        double trueValue = feature.mean;
        double[] falseCache = feature.execute(falseFiles);
        double falseValue = feature.mean;

        double threshold = (trueValue + falseValue) / 2;

        double value = calculateHits(trueCache, falseCache, threshold);

        String code = feature.toJava();
        double percentage = value / (trueCache.length + falseCache.length);
        System.out.println("hits: " + value + "(" + percentage + ")");
        System.out.println(code);
    }

    private int calculateHits(double[] trueCache, double[] falseCache, double threshold) {

        int hits = 0;
        int misses = 0;

        for (int i = 0; i < trueCache.length; i++) {
            if (trueCache[i] > threshold) hits++;
            else misses++;
        }

        for (int i = 0; i < falseCache.length; i++) {
            if (falseCache[i] < threshold) hits++;
            else misses++;
        }

        return Math.max(hits, misses);

    }

    class HaarFeature implements Comparable {

        private int mode, x, y, width, height, adjacency, shape;

        private int hits;

        public HaarFeature(int mode, int x, int y, int width, int height, int adjacency, int shape) {
            this.mode = mode;
            this.x = x;
            this.y = y;
            this.width = width;
            this.height = height;
            this.adjacency = adjacency;
            this.shape = shape;
        }

        public int getHits() {
            return hits;
        }

        public void setHits(int hits) {
            this.hits = hits;
        }

        public String toJava() {

            switch (mode) {
                case ONE_RECT:
                    return "image.getOneRectangleFeature(" + x + "," + y + ", " + width + ", " + height + ")";
                case TWO_RECT:
                    return "image.getTwoRectangleFeature(" + x + "," + y + ", " + width + ", " + height + ", " + adjacency + ", " + shape + ")";
                case THREE_RECT:
                    return "image.getThreeRectangleFeature(" + x + "," + y + ", " + width + ", " + height + ", " + adjacency + ")";
                default:
                    return "image.getFourRectangleFeature(" + x + "," + y + ", " + width + ", " + height + ", " + adjacency + ")";
            }

        }

        public double execute(HaarRegions image) {
            switch (mode) {
                case ONE_RECT:
                    return image.getOneRectangleFeature(x, y, width, height);
                case TWO_RECT:
                    return image.getTwoRectangleFeature(x, y, width, height, adjacency, shape);
                case THREE_RECT:
                    return image.getThreeRectangleFeature(x, y, width, height, adjacency);
                default:
                    return image.getFourRectangleFeature(x, y, width, height, adjacency);
            }
        }

        private double mean;

        public double[] execute(Vector<HaarRegions> images) {

            double[] cached = new double[images.size()];

            mean = 0;

            for (int i = 0; i < images.size(); i++) {
                HaarRegions image = images.elementAt(i);
                double value = execute(image);
                mean += value;
                cached[i] = value;
            }

            mean = mean / images.size();

            return cached;

        }

        public double estimateSimilarity(HaarFeature other) {

            if (other.mode != this.mode) return Double.MAX_VALUE;

            if (this.mode > ONE_RECT && other.adjacency != this.adjacency) return Double.MAX_VALUE;

            double value = 0;
            value += euclid(this.x, other.x);
            value += euclid(this.y, other.y);
            value += euclid(this.width, other.width);
            value += euclid(this.height, other.height);
            return Math.sqrt(value);
        }

        private double euclid(double value1, double value2) {
            double diff = value1 - value2;
            return (diff * diff);
        }


        public int compareTo(Object o) {
            HaarFeature other = (HaarFeature) o;
            if (other.hits == hits) return 0;
            if (other.hits > hits) return 1;
            return -1;
        }

        public String toString() {
            return "HF, Hits=" + hits + ", Java=" + toJava();
        }
    }

}
