package problems.face;

import ga.core.GAIndividual;
import problems.face.genes.*;
import ac.essex.ooechs.imaging.commons.HaarRegions;

/**
 * This individual defines the parameters for a Haar Imaging function.
 *
 * @author Olly Oechsle, University of Essex, Date: 21-May-2007
 * @version 1.0
 */
public class FaceDetectorIndividual extends GAIndividual {

    public static final int ONE_RECT = 1;
    public static final int TWO_RECT = 2;
    public static final int THREE_RECT = 3;
    public static final int FOUR_RECT = 4;

    public int mode;

    public int x, y;

    public FaceDetectorIndividual(int x, int y, int mode) {
        super(4);
        this.x = x;
        this.y = y;
        this.mode = mode;
        genes[0] = new WidthGene();
        genes[1] = new HeightGene();
        genes[2] = new AdjacencyGene();
        genes[3] = new ShapeGene();
    }

    public double execute(Object o) {

        //int x = (int) genes[0].value;
        //int y = (int) genes[1].value;
        int width = (int) genes[0].value;
        int height = (int) genes[1].value;
        int adjacency = (int) genes[2].value;
        int shape = (int) genes[3].value;

        HaarRegions image = (HaarRegions) o;

        switch (mode) {
            case ONE_RECT:
                return image.getOneRectangleFeature(x, y, width, height);
            case TWO_RECT:
                return image.getTwoRectangleFeature(x, y, width, height, adjacency, shape);
            case THREE_RECT:
                return image.getThreeRectangleFeature(x, y, width, height, adjacency);
            default:
                return image.getFourRectangleFeature(x, y, width, height, adjacency);
        }

    }

    public String toJava() {

/*        int x = (int) genes[0].value;
        int y = (int) genes[1].value;*/
        int width = (int) genes[0].value;
        int height = (int) genes[1].value;
        int adjacency = (int) genes[2].value;
        int shape = (int) genes[3].value;

        switch (mode) {
            case ONE_RECT:
                return "image.getOneRectangleFeature(" + x + "," + y + ", " +  width + ", " +  height + ")";
            case TWO_RECT:
                return "image.getTwoRectangleFeature(" + x + "," + y + ", " +  width + ", " +  height + ", " + adjacency + ", " + shape + ")";
            case THREE_RECT:
                return "image.getThreeRectangleFeature(" + x + "," + y + ", " +  width + ", " +  height + ", " + adjacency + ")";
            default:
                return "image.getFourRectangleFeature(" + x + "," + y + ", " +  width + ", " +  height + ", " + adjacency + ")";
        }

    }
    
}
