package problems.jasmine;

import ga.core.GAIndividual;
import ga.genes.SmallIntGene;
import ga.genes.BooleanGene;
import ga.genes.TinyDoubleGene;
import ga.genes.PercentageGene;
import ac.essex.ooechs.imaging.shapes.ExtraShapeData;
import ac.essex.ooechs.imaging.shapes.SegmentedShape;

/**
 * Shape individual can try a number of different shape techniques.
 *
 * @author Olly Oechsle, University of Essex, Date: 22-May-2007
 * @version 1.0
 */
public class ShapeIndividual extends GAIndividual {

    public static final int CORNERS = 1;
    public static final int COUNTHOLLOWS = 2;
    public static final int BALANCEX = 3;
    public static final int BALANCEY = 4;
    public static final int DENSITY = 5;
    public static final int ASPECTRATIO = 6;
    public static final int JOINTS = 7;
    public static final int ENDS = 8;
    public static final int ROUNDNESS = 9;
    public static final int BALANCEXENDS = 10;
    public static final int BALANCEYENDS = 11;
    public static final int CLOSESTENDTOCOG = 12;
    public static final int CLOSESTPIXELTOCOG = 13;
    public static final int HORIZONTAL_SYMMETRY = 14;
    public static final int VERICAL_SYMMETRY = 15;
    public static final int INVERSE_HORIZONTAL_SYMMETRY = 16;
    public static final int INVERSE_VERTICAL_SYMMETRY = 17;

    protected int mode;

    public ShapeIndividual(int mode) {

        super(2);
        this.mode = mode;

        // which threshold to use
        switch (mode) {
            case CORNERS:
            case COUNTHOLLOWS:
            case JOINTS:
            case ENDS:
                genes[0] = new SmallIntGene();
                break;
            case ASPECTRATIO:
            case BALANCEX:
            case BALANCEY:
            case ROUNDNESS:
                genes[0] = new TinyDoubleGene();
                break;
            case BALANCEXENDS:
            case BALANCEYENDS:
            case CLOSESTENDTOCOG:
            case CLOSESTPIXELTOCOG:
            case DENSITY:
            case HORIZONTAL_SYMMETRY:
            case VERICAL_SYMMETRY:
            case INVERSE_HORIZONTAL_SYMMETRY:
            case INVERSE_VERTICAL_SYMMETRY:
                genes[0] = new PercentageGene();
                break;
            default:
                throw new RuntimeException("Invalid mode: " + mode);

        }
        genes[1] = new BooleanGene(); // polarity

    }


    public int classify(Object o) {
        ExtraShapeData s = (ExtraShapeData) o;
        return (int) execute(new ShapeData(s.getShape(), s));
    }

    public double execute(Object o) {
        ShapeData data = (ShapeData) o;

        if (genes[1].value == 1.0) {
            return get(data.shape, data.extra, mode) > genes[0].value? 1 : 0;
        } else {
            return get(data.shape, data.extra, mode) < genes[0].value? 1 : 0;
        }

    }

    public static double get(SegmentedShape s, int mode) {
        return get(s, new ExtraShapeData(s), mode);
    }

    public static double get(SegmentedShape shape, ExtraShapeData extra, int mode) {

        switch (mode) {
            case CORNERS:
                return extra.countCorners();
            case COUNTHOLLOWS:
                return extra.countHollows();
            case JOINTS:
                return extra.getJoints();
            case ENDS:
                return extra.getEnds();
            case ASPECTRATIO:
                return extra.getAspectRatio();
            case BALANCEX:
                return extra.getBalanceX();
            case BALANCEY:
                return extra.getBalanceY();
            case ROUNDNESS:
                return extra.getRoundness();
            case BALANCEXENDS:
                return extra.getEndBalanceX();
            case BALANCEYENDS:
                return extra.getEndBalanceY();
            case CLOSESTENDTOCOG:
                return extra.getClosestEndToCog();
            case CLOSESTPIXELTOCOG:
                return extra.getClosestPixelToCog();
            case DENSITY:
                return extra.getDensity();
            case HORIZONTAL_SYMMETRY:
                return extra.getHorizontalSymmetry();
            case VERICAL_SYMMETRY:
                return extra.getVerticalSymmetry();
            case INVERSE_HORIZONTAL_SYMMETRY:
                return extra.getInverseHorizontalSymmetry();
            case INVERSE_VERTICAL_SYMMETRY:
                return extra.getInverseVerticalSymmetry();
            default:
                throw new RuntimeException("Invalid mode: " + mode);
        }
    }

    public String toJava() {
        double threshold = genes[0].value;
        if (genes[1].value == 1.0) {
            return "ShapeIndividual.get(shape, " + mode + ") > " + threshold + "? 1 : 0";
        } else {
            return "ShapeIndividual.get(shape, " + mode + ") < " + threshold + "? 1 : 0";
        }
    }
    
}
